/******************************************************************************* 
	Project: The Economics of Women Entrepreneurship
	Paper: Economic Inclusion through Entrepreneurship: Evidence on Building 
		   Skills for Women
	Title: Two Event Studies for Covid-19
	Authors: Megan Lang and Julia Seither
	RA: Maria Sofia Casabianca

	Description: Appendix figures  
					
	STATA version: 18

	Date: Jan 19 - 2024 (Last Update)										 
*******************************************************************************/

********************************************************************************
**************************** SMS DATA PREPARATION ******************************
********************************************************************************
clear all
set more off
set varabbrev off

use "$path/../Data/IE_SMSFigure_Biweekly.dta", clear

gen treat = 1 if tmt == "B"
replace treat = 0 if tmt == "C"

gen treatxcovid1 = treat*covid1

sort yrmowe
encode yrmowe, gen(date)

rename event_date sbs_event_date

// CREATING EVENT DATE WITH PERIOD 0 IS ONE PERIOD BEFORE THE FIRST COVID LOCKDOWN

gen covid_event_date = date - 38
tab covid_event_date, generate(t_)

local new_name "m37	m36	m35	m34	m33	m32	m31	m30	m29	m28	m27	m26	m25	m24	m23	m22	m21	m20	m19	m18	m17	m16	m15	m14	m13	m12	m11	m10	m9	m8	m7	m6	m5	m4	m3	m2	m1	p0	p1	p2	p3	p4	p5	p6	p7	p8	p9	p10	p11	p12	p13	p14	p15	p16	p17	p18	p19	p20	p21	p22	p23	p24	p25	p26	p27	p28	p29	p30	p31	p32	p33	p34	p35	p36	p37	p38	p39	p40	p41"

local old_name "t_1	t_2	t_3	t_4	t_5	t_6	t_7	t_8	t_9	t_10 t_11 t_12 t_13 t_14 t_15 t_16 t_17 t_18 t_19 t_20 t_21 t_22 t_23 t_24 t_25 t_26 t_27 t_28 t_29 t_30 t_31 t_32 t_33 t_34 t_35 t_36 t_37 t_38 t_39 t_40 t_41 t_42 t_43 t_44 t_45 t_46 t_47 t_48 t_49 t_50 t_51 t_52 t_53 t_54 t_55 t_56 t_57 t_58 t_59 t_60 t_61 t_62 t_63 t_64 t_65 t_66 t_67 t_68 t_69 t_70 t_71 t_72 t_73 t_74 t_75 t_76 t_77 t_78 t_79"

rename (`old_name') (`new_name')


global dummies2 "m10 m9 m8 m7 m6 m5 m4 m3 m2 m1 p1 p2 p3 p4 p5 p6 p7 p8 p9 p10 p11 p12 p13 p14 p15 p16 p17 p18 p19 p20 p21"

foreach i in $dummies2{
	gen treatx`i' = treat*`i'
	label variable `i' "`i'"
}

global interactions2 "treatxm10 treatxm9 treatxm8 treatxm7 treatxm6 treatxm5 treatxm4 treatxm3 treatxm2 treatxm1 treatxp1 treatxp2 treatxp3 treatxp4 treatxp5 treatxp6 treatxp7 treatxp8 treatxp9 treatxp10 treatxp11 treatxp12 treatxp13 treatxp14 treatxp15 treatxp16 treatxp17 treatxp18 treatxp19 treatxp20 treatxp21"

// WITH INTERACTIONS BY TREATMENT: From -10 to 21
eststo: reghdfe Sales_Clean_w $dummies2 $interactions2 [aw=sms_weight], absorb(id) vce(cluster id)
	matrix M = r(table)
	matselrc M Coef_M, r(b ll ul)
	matrix Coef_M = Coef_M'
		matrix Coef_M[63,1] = 0
		matrix Coef_M[63,2] = .
		matrix Coef_M[63,3] = .
	matrix x = (-9.8\ -8.8\ -7.8\ -6.8\ -5.8\ -4.8\ -3.8\ -2.8\ -1.8\ -0.8\ 1.2\ 2.2\ 3.2\ 4.2\ 5.2\ 6.2\ 7.2\ 8.2\ 9.2\ 10.2\ 11.2\ 12.2\ 13.2\ 14.2\ 15.2\ 16.2\ 17.2\ 18.2\ 19.2\ 20.2\ 21.2\ -10\ -9\ -8\ -7\ -6\ -5\ -4\ -3\ -2\ -1\ 1\ 2\ 3\ 4\ 5\ 6\ 7\ 8\ 9\ 10\ 11\ 12\ 13\ 14\ 15\ 16\ 17\ 18\ 19\ 20\ 21\ 0 )
	matrix Coef_M = Coef_M, x

preserve
clear
svmat2 Coef_M, names(col) rnames(varname)

rename c1 xvar

gen N =_n
gen treat = 0
	replace treat = 1 if N > 31
	replace treat = 0 if xvar == 0

twoway (scatter b xvar if treat == 1, color(edkblue)) (rcap ll ul xvar if treat == 1, color(black%50)) ///
	   (scatter b xvar if treat == 0, color(stred)) (rcap ll ul xvar if treat == 0, color(dimgrey) lp(dash)), ///
	   ytitle("Average Bi-weekly Sales Revenues") xlabel(#37, angle(45) labsize(vsmall)) ///
	   legend(order(1 "Treated" 2 "CI - Treated" 3 "Control" 4 "CI - Control") region(style(none)) pos(6) r(2))  ///
	   graphregion(color(white)) name(plot4, replace) xtitle("Normalized Bi-weekly Period") ///
	   xline(1, lp(solid) lcolor(red%40)) yline(0, lcolor(black%70)lp(solid)) xline(12, lp(solid) lcolor(red%40)) yline(0, lcolor(black%70)lp(solid))
graph export "$path/../Output/Figures/Covid-19_EventStudy_weighted.jpg", replace as(jpg) name("plot4") quality(100)
	
restore	

est clear